from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zbl
    from typing import Self, Sequence, Literal, Callable, TypeVar, Generator
    from ._typing import _Area, _Areas
    from aclib.cv._typing import HexColorRange
    T = TypeVar('T')

import sys

from aclib.cv import Image, FontLib, DotsetLib, TargetList
from aclib.winlib import winapi

from ._typing import WinTarget
from .window import Window


class CvWindow(Window):

    def __init__(self):
        super().__init__()
        self.__cvflib = FontLib()
        self.__cvdlib = DotsetLib()
        self.__zblcapture: zbl.Capture|None = None

    def __del__(self):
        self.cv_stop_wgc()

    FontLib = FontLib
    DotsetLib = DotsetLib

    def cvset(self, dlib: DotsetLib|None=..., flib: FontLib|None=...) -> Self:
        """传入None删除已设置的库，不传(或传入...)则不修改"""
        if isinstance(flib, FontLib):
            self.__cvflib = flib
        if isinstance(dlib, DotsetLib):
            self.__cvdlib = dlib
        if flib is None:
            self.__cvflib = FontLib()
        if dlib is None:
            self.__cvdlib = DotsetLib()
        return self

    def cv_start_wgc(self):
        """
        启用wgc截图，需要 ***使用64位python下载zbl依赖库：`pip install zbl`***  
        - 调用此方法可对使用gdi方式截图无效的窗口提供截图支持，但仅支持对`overlapped window`顶层窗口进行截图  
        - 不支持的窗口(`child window / popup window`)调用此方法会报参数错误  
        - 在`cv_stop_wgc`被调用前用户会看到窗口被黄色边框包裹提示用户该窗口正在被录制  
        """
        if sys.maxsize <= 2**32:
            raise RuntimeError(
                '要启用wgc截图，请使用64位Python')
        if not self.__zblcapture:
            import zbl
            self.__zblcapture = zbl.Capture(window_handle=self.handle)
            self.__zblcapture.__enter__()

    def cv_stop_wgc(self):
        if self.__zblcapture:
            self.__zblcapture.__exit__()
            self.__zblcapture = None

    def cvcapture(self, area: _Area=None, savepath='') -> Image | None:
        if area and not isinstance(area[0], int):
            area = tuple(map(int, area))
        if self.__zblcapture:
            screenshot = Image.fromarray(self.__zblcapture.grab())
            if area: screenshot = screenshot.crop(area[:2], area[2:])
        else:
            size, buffer = winapi.CaptureWindow(self.handle, area)
            screenshot = Image.frombuffer(buffer, size)
        if savepath:
            screenshot.tofile(savepath)
        return screenshot


    def __cviter_target(self, capareas: _Area|_Areas|None, capscale: float, func: Callable[[Image], T]) -> Generator[tuple[tuple[int,int], T], None, None]:
        if capareas is None:
            capareas = [(0, 0, *self.clientsize)]
        if hasattr(capareas[0], '__index__'):
            capareas = [capareas]
        for caparea in capareas:
            capstart = caparea[:2]
            screenshot = self.cvcapture(caparea)
            if not screenshot:
                continue
            if capscale != 1:
                screenshot = screenshot.scale(capscale)
            yield capstart, func(screenshot)

    def __cvfind_target(self, capareas: _Area|_Areas|None, capscale: float, func: Callable[[Image], WinTarget]) -> WinTarget:
        for start, target in self.__cviter_target(capareas, capscale, func):
            if target:
                return WinTarget.fromtarget(target, self).scale(1/capscale, (0, 0)).offset(*start)
        return WinTarget()

    def __cvfind_targets(self, capareas: _Area|_Areas|None, capscale: float, func: Callable[[Image], TargetList]) -> list[WinTarget]:
        found = []
        for start, targets in self.__cviter_target(capareas, capscale, func):
            found.extend(WinTarget.fromtarget(target, self).scale(1/capscale, (0, 0)).offset(*start) for target in targets)
        return found


    def cvfindcolor(self, rgbranges: HexColorRange, areas: _Area|_Areas=None) -> WinTarget:
        return self.__cvfind_target(areas, 1, lambda cap: cap.findcolor(rgbranges))

    def cvfindcolors(self, rgbranges: HexColorRange, areas: _Area|_Areas=None) -> list[WinTarget]:
        return self.__cvfind_targets(areas, 1, lambda cap: cap.findcolors(rgbranges))

    def cvfinddotset(self,
        dotsets: str|Sequence[str],
        areas: _Area|_Areas=None,
        matchcolor: Literal[0,1]|HexColorRange=None,
        similarity = 0.9,
        capscale = 1.0,
        usedlib: DotsetLib|None=None,
    ) -> WinTarget:
        if len(_dlib := usedlib or self.__cvdlib) == 0:
            return WinTarget()
        return self.__cvfind_target(
            areas, capscale, lambda cap: cap.finddotset(_dlib, dotsets, matchcolor, similarity))

    def cvfinddotsets(self,
        dotsets: str|Sequence[str],
        areas: _Area|_Areas=None,
        matchcolor: Literal[0,1]|HexColorRange=None,
        similarity = 0.9,
        capscale = 1.0,
        ignore_overlaps = False,
        usedlib: DotsetLib|None=None,
    ) -> list[WinTarget]:
        if len(_dlib := usedlib or self.__cvdlib) == 0:
            return []
        return self.__cvfind_targets(
            areas, capscale, lambda cap: cap.finddotsets(_dlib, dotsets, matchcolor, similarity, 1, ignore_overlaps))

    def cvfindtext(self,
        texts: str|Sequence[str],
        areas: _Area|_Areas=None,
        matchcolor: Literal[0,1]|HexColorRange=None,
        similarity = 0.9,
        txtdir: Literal[0,1]=0,
        txtwrap = True,
        capscale = 1.0,
        charset: str|Literal['']=None,
        useflib: FontLib|None=None,
    ) -> WinTarget:
        if len(_flib := useflib or self.__cvflib) == 0:
            return WinTarget()
        return self.__cvfind_target(
            areas, capscale, lambda cap: cap.findtext(_flib, texts, matchcolor, similarity, txtdir, txtwrap, 1, charset))

    def cvfindtexts(self,
        texts: str|Sequence[str],
        areas: _Area|_Areas=None,
        matchcolor: Literal[0,1]|HexColorRange=None,
        similarity = 0.9,
        txtdir: Literal[0,1]=0,
        txtwrap = True,
        capscale = 1.0,
        charset: str|Literal['']=None,
        ignore_overlaps = False,
        useflib: FontLib|None=None,
    ) -> list[WinTarget]:
        if len(_flib := useflib or self.__cvflib) == 0:
            return []
        return self.__cvfind_targets(
            areas, capscale, lambda cap: cap.findtexts(_flib, texts, matchcolor, similarity, txtdir, txtwrap, 1, charset, ignore_overlaps))

    def cvocr(self,
        areas: _Area|_Areas=None,
        matchcolor: Literal[0,1]|HexColorRange=None,
        similarity = 0.9,
        txtdir: Literal[0,1]=0,
        txtwrap = True,
        capscale = 1.0,
        charset: str|Literal['']=None,
        useflib: FontLib|None=None,
    ) -> list[WinTarget]:
        if len(_flib := useflib or self.__cvflib) == 0:
            return []
        texts = []
        for start, ocrgroups in self.__cviter_target(
            areas, capscale, lambda cap: cap.ocr(_flib, matchcolor, similarity, txtdir, txtwrap, 1, charset)
        ):
            texts.extend(ocrgroup.join().scale(1/capscale, (0, 0)).offset(*start) for ocrgroup in ocrgroups)
        return texts



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self
    from aclib.winlib import wintype
    from ._typing import _Pos, _Size, _Area

import time, getpass
from aclib.winlib.winapi import *
from aclib.winlib.wincon import *


__all__ = [
    'BaseWindow'
]


class BaseWindow(object):

    def __init__(self):
        self.__handle = 0

    def _sethandle_(self, hwnd: int) -> Self:
        self.__handle = hwnd
        return self

    def __bool__(self):
        return bool(self.handle)

    def __repr__(self):
        summary = self.title.replace('\n', '\\n') or self.classname
        return f'<{self.__class__.__name__}-{self.handle} {summary}>'


    @property
    def handle(self) -> int:
        return self.__handle

    @property
    def classname(self) -> str:
        return GetWindowClassName(self.handle)

    @property
    def title(self) -> str:
        return GetWindowTitle(self.handle)

    def settitle(self, title: str) -> bool:
        return SetWindowTitle(self.handle, title)


    @property
    def threadid(self) -> int:
        return GetWindowThreadProcessId(self.handle)[0]

    @property
    def processid(self) -> int:
        return GetWindowThreadProcessId(self.handle)[1]

    @property
    def threadprocessid(self) -> tuple[int, int]:
        return GetWindowThreadProcessId(self.handle)

    @property
    def creationtime(self) -> int:
        hthread = OpenThreadHandle(self.threadid)
        ctime = GetThreadTimes(hthread)[0]
        CloseHandle(hthread)
        return ctime

    @property
    def exittime(self) -> int:
        hthread = OpenThreadHandle(self.threadid)
        etime = GetThreadTimes(hthread)[1]
        CloseHandle(hthread)
        return etime


    def tolayeredwindow(self):
        LayerWindow(self.handle)

    def tounlayeredwindow(self):
        UnlayerWindow(self.handle)


    @property
    def isexistent(self) -> bool:
        return IsWindowExistent(self.handle)

    @property
    def isenabled(self) -> bool:
        return IsWindowEnabled(self.handle)

    @property
    def isvisible(self) -> bool:
        return IsWindowVisible(self.handle)

    @property
    def isviewable(self) -> bool:
        (l, t, r, b), (sw, sh) = self.rect.values(), GetScreenSize()
        return l < sw and t < sh and r > 0 and b > 0

    @property
    def ismaximized(self) -> bool:
        return IsWindowMaximized(self.handle)

    @property
    def isminimized(self) -> bool:
        return IsWindowMinimized(self.handle)

    @property
    def isnormalized(self) -> bool:
        return IsWindowNormalized(self.handle)

    @property
    def isforeground(self) -> bool:
        return self.handle == GetForegroundWindow()

    @property
    def istopmost(self) -> bool:
        return IsWindowTopMost(self.handle)


    def close(self):
        CloseWindow(self.handle)

    def destroy(self):
        DestroyWindow(self.handle)

    def enable(self):
        EnableWindow(self.handle)

    def disable(self):
        DisableWindow(self.handle)

    def show(self):
        ShowWindow(self.handle)

    def hide(self):
        HideWindow(self.handle)

    def maximize(self):
        MaximizeWindow(self.handle)

    def minimize(self):
        MinimizeWindow(self.handle)

    def normalieze(self):
        NormalizeWindow(self.handle)

    def settopmost(self, topmost: bool):
        SetWindowTopMost(self.handle, topmost)


    @property
    def rect(self) -> wintype.RECT:
        return GetWindowRect(self.handle)

    @property
    def rectR(self) -> wintype.RECT:
        return GetWindowRectR(self.handle)

    @property
    def pos(self) -> _Pos:
        return self.rect.start

    @property
    def posR(self) -> _Pos:
        return self.rectR.start

    @property
    def size(self) -> _Size:
        return self.rect.size

    def setpossize(self, newpos: _Pos, newsize: _Size):
        SetWindowPosSize(self.handle, newpos, newsize)

    def setpossizeR(self, newposR: _Pos, newsize: _Size):
        SetWindowPosSizeR(self.handle, newposR, newsize)

    def setpos(self, newpos: _Pos):
        SetWindowPosSize(self.handle, newpos)

    def setposR(self, newposR: _Pos):
        SetWindowPosSizeR(self.handle, newposR)

    def setsize(self, newsize: _Size):
        SetWindowPosSize(self.handle, None, newsize)


    @property
    def clientrect(self) -> wintype.RECT:
        return GetClientRect(self.handle)

    @property
    def clientrectR(self) -> wintype.RECT:
        return GetClientRectR(self.handle)

    @property
    def clientpos(self) -> _Pos:
        return self.clientrect.start

    @property
    def clientposR(self) -> _Pos:
        return self.clientrectR.start

    @property
    def clientsize(self) -> _Size:
        return self.clientrect.size

    def setclientpossize(self, newclientpos: _Pos, newclientsize: _Size):
        SetClientPosSize(self.handle, newclientpos, newclientsize)

    def setclientpossizeR(self, newclientposR: _Pos, newclientsize: _Size):
        SetClientPosSizeR(self.handle, newclientposR, newclientsize)

    def setclientpos(self, newclientpos: _Pos):
        SetClientPosSize(self.handle, newclientpos)

    def setclientposR(self, newclientposR: _Pos):
        SetClientPosSizeR(self.handle, newclientposR)

    def setclientsize(self, newclientsize: _Size):
        SetClientPosSize(self.handle, None, newclientsize)


    @property
    def transparency(self) -> float:
        return GetWindowTransparency(self.handle)

    def settransparency(self, transparency: float):
        SetWindowTransparency(self.handle, transparency)


    def getcolor(self, pos: _Pos) -> int:
        """get decimal bgr color at given pos"""
        return GetPixel(self.handle, pos)

    def capture(self, area: _Area) -> tuple[_Size, bytearray]:
        return CaptureWindow(self.handle, area)


    def movemouse(self, topos: _Pos):
        PostMessage(self.handle, WM_MOUSEMOVE, 0, MakeLong(*topos))


    def leftdown(self, pos: _Pos):
        PostMessage(self.handle, WM_LBUTTONDOWN, MK_LBUTTON, MakeLong(*pos))

    def leftup(self, pos: _Pos):
        PostMessage(self.handle, WM_LBUTTONUP, MK_LBUTTON, MakeLong(*pos))

    def leftclick(self, pos: _Pos):
        self.leftdown(pos)
        self.leftup(pos)

    def leftdbclick(self, pos: _Pos):
        self.leftclick(pos)
        self.leftclick(pos)

    def leftdrag(self, pos: _Pos, topos: _Pos):
        self.leftdown(pos)
        self.movemouse(topos)
        self.leftup(topos)

    def leftdragR(self, pos: _Pos, dpos: _Pos):
        self.leftdrag(pos, (pos[0] + dpos[0], pos[1] + dpos[1]))


    def rightdown(self, pos: _Pos):
        PostMessage(self.handle, WM_RBUTTONDOWN, MK_RBUTTON, MakeLong(*pos))

    def rightup(self, pos: _Pos):
        PostMessage(self.handle, WM_RBUTTONUP, MK_RBUTTON, MakeLong(*pos))

    def rightclick(self, pos: _Pos):
        self.rightdown(pos)
        self.rightup(pos)

    def rightdbclick(self, pos: _Pos):
        self.rightclick(pos)
        self.rightclick(pos)

    def rightdrag(self, pos: _Pos, topos: _Pos):
        self.rightdown(pos)
        self.movemouse(topos)
        self.rightup(topos)

    def rightdragR(self, pos: _Pos, dpos: _Pos):
        self.rightdrag(pos, (pos[0] + dpos[0], pos[1] + dpos[1]))


    def middown(self, pos: _Pos):
        PostMessage(self.handle, WM_MBUTTONDOWN, MK_MBUTTON, MakeLong(*pos))

    def midup(self, pos: _Pos):
        PostMessage(self.handle, WM_MBUTTONUP, MK_MBUTTON, MakeLong(*pos))

    def midclick(self, pos: _Pos):
        self.middown(pos)
        self.midup(pos)

    def middrag(self, pos: _Pos, topos: _Pos):
        self.middown(pos)
        self.movemouse(topos)
        self.midup(topos)

    def middragR(self, pos: _Pos, dpos: _Pos):
        self.middrag(pos, (pos[0] + dpos[0], pos[1] + dpos[1]))

    def wheelup(self, pos, times=1):
        self.movemouse(pos)
        lParam = MakeLong(*ClientToScreen(self.handle, *pos))
        for i in range(times):
            PostMessage(self.handle, WM_MOUSEWHEEL, WHEEL_DELTA, lParam)

    def wheeldown(self, pos, times=1):
        self.movemouse(pos)
        lParam = MakeLong(*ClientToScreen(self.handle, *pos))
        for i in range(times):
            PostMessage(self.handle, WM_MOUSEWHEEL, -WHEEL_DELTA, lParam)


    def keydown(self, key: int | str):
        PostMessage(self.handle, WM_IME_KEYDOWN, *MakeKeyMessageParam(key)[0])

    def keyup(self, key:int|str):
        PostMessage(self.handle, WM_IME_KEYUP, *MakeKeyMessageParam(key)[1])

    def keypress(self, key: int | str):
        self.keydown(key)
        self.keyup(key)

    def waitkey(self, *keys: int | str, msg='', cmdmode=False) -> int | str:
        vkmap = {}
        for key in keys:
            vkcode = MakeKeyCode(key)
            if not vkcode:
                raise ValueError(f'invalid key <- {key} ->')
            vkmap[vkcode] =  key
        vkmap = vkmap or VK_MAP_KEYBOARD
        print(msg)
        while True:
            for vkcode in vkmap:
                if GetAsyncKeyState(vkcode):
                    if cmdmode:
                        if vkcode != 13: self.keypress('enter')
                        getpass.getpass('')
                    return vkmap[vkcode]
            time.sleep(.001)

    def sendtext(self, text: str):
        for char in text:
            PostMessage(self.handle, WM_CHAR, ord(char), 0)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self, Literal
    from ._typing import _Area, _Areas

import os, itertools
from aclib.dm import DM, DmDotsetLib

from ._typing import WinTarget
from .window import Window


class DmWindow(Window):

    def __init__(self):
        super().__init__()
        self.__dmbind = False
        self.__dmwordlock = False
        self.__dmdlib: DmDotsetLib = None

    @property
    def __dm(self) -> DM:
        del DmWindow.__dm
        self.__dm = DM()
        self.__dm.SetShowErrorMsg(0)
        return self.__dm

    def dmset(self,
        displaymode: Literal['dx', 'dx2', 'dx3', 'gdi', 'gdi2'] | None = '',
        flib: str | None = '',
        dlib: str | None = '',
        dlibscale: float = 1.0
    ) -> Self:
        """传入空字符串时对应参数不做更改，传入None时清除已设置的值"""
        if displaymode:
            self.__dm.UnBindWindow()
            self.__dm.BindWindow(self.handle, displaymode, 'windows', 'windows', 0)
            self.__dmbind = True
        if flib:
            if not os.path.isfile(flib):
                raise FileNotFoundError(
                    f'No such file or directory: {flib}')
            self.__dm.SetDict(0, flib)
        if dlib:
            self.__dmdlib = DmDotsetLib.fromfile(dlib).scale(dlibscale)
        if displaymode is None:
            self.__dm.UnBindWindow()
            self.__dmbind = False
        if flib is None:
            self.__dm.ClearDict()
            self.__dmwordlock = False
        if dlib is None:
            self.__dmdlib = None
        return self


    def __dmassert_bind(self):
        assert self.__dmbind, 'please call dmset() first'

    def __dmtransfer_areas(self, areas: _Area|_Areas|None) -> _Areas:
        self.__dmassert_bind()
        if areas is None:
            return [(0, 0, *self.clientsize)]
        if hasattr(areas[0], '__index__'):
            return [areas]
        return areas


    def dmcapture(self, area: _Area=None, savepath: str=''):
        self.__dmassert_bind()
        area = area or (0, 0, *self.clientsize)
        self.__dm.Capture(*area, savepath)

    def dmfindcolor(self, color: str, areas: _Area|_Areas=None) -> WinTarget:
        areas = self.__dmtransfer_areas(areas)
        for area, color in itertools.product(areas, color.split('|')):
            x, y, success = self.__dm.FindColor(*area, color, 1, 0)
            if success:
                return WinTarget(x, y, x+1, y+1, color, 1.0, self)
        return WinTarget()

    def dmfindcolors(self, color: str, areas: _Area|_Areas=None) -> list[WinTarget]:
        found, areas = [], self.__dmtransfer_areas(areas)
        for area, color in itertools.product(areas, color.split('|')):
            res = self.__dm.FindColorEx(*area, color, 1, 0)
            if not res: continue
            xs, ys = [[int(coor) for coor in coors.split(',')] for coors in res.split('|')]
            for i in range(min(len(xs), len(ys))):
                x, y = xs[i], ys[i]
                found.append(WinTarget(x, y, x+1, y+1, color, 1.0, self))
        return found

    def dmfindcolorblock(self, colorBlock: str, areas: _Area|_Areas=None, similarity=1, scale=1) -> WinTarget:
        """ colorBlock: '{w}x{h}x{color}|{w}x{h}x{color}' like '3x5xefefef-101010' """
        areas = self.__dmtransfer_areas(areas)
        for area, colorBlock in itertools.product(areas, colorBlock.split('|')):
            w,h,color = [(round(int(w)*scale), round(int(h)*scale), color) for w,h,color in [(info for info in colorBlock.split('x'))]][0]
            dmColorBlock = ','.join([f'{i%w}|{i//w}|{color}' for i in range(w*h)])
            x, y, success = self.__dm.FindMultiColor(*area, color, dmColorBlock, similarity, 0)
            if success:
                return WinTarget(x, y, x+w, y+h, f'{w}x{h}x{color}', similarity, self)
        return WinTarget()

    def dmfindcolorblocks(self, colorBlock: str, areas: _Area|_Areas=None, similarity=1, scale=1) -> list[WinTarget]:
        """ colorBlock: '{w}x{h}x{color}|{w}x{h}x{color}' like '3x5xefefef-101010' """
        found, areas = [], self.__dmtransfer_areas(areas)
        for area, colorBlock in itertools.product(areas, colorBlock.split('|')):
            w,h,color = [(round(int(w)*scale), round(int(h)*scale), color) for w,h,color in [(info for info in colorBlock.split('x'))]][0]
            dmColorBlock = ','.join([f'{i%w}|{i//w}|{color}' for i in range(w*h)])
            res = self.__dm.FindMultiColorEx(*area, color, dmColorBlock, similarity, 0)
            if not res: continue
            xs, ys = [[int(coor) for coor in coors.split(',')] for coors in res.split('|')]
            for i in range(min(len(xs), len(ys))):
                found.append(WinTarget(xs[i], ys[i], xs[i]+w, ys[i]+h, f'{w}x{h}x{color}', similarity, self))
        return found

    def dmfinddotset(self, dotsetname: str, areas: _Area|_Areas=None, color: str=None, similarity=1, scale=1) -> WinTarget:
        if not self.__dmdlib:
            return WinTarget()
        areas = self.__dmtransfer_areas(areas)
        for area, dotsetname in itertools.product(areas, dotsetname.split('|')):
            for dotset in self.__dmdlib.group(dotsetname):
                dotset = dotset.scale(scale).asmatchcolor(color)
                x, y, success = self.__dm.FindMultiColor(*area, dotset.matchcolor, dotset.tmpl, similarity, 0)
                if success:
                    start = dotset.getrealpos((x, y))
                    end = start[0] + dotset.width, start[1] + dotset.height
                    return WinTarget(*start, *end, dotsetname, similarity, self)
        return WinTarget()

    def dmfinddotsets(self, dotsetname: str, areas: _Area|_Areas=None, color: str=None, similarity=1, scale=1) -> list[WinTarget]:
        if not self.__dmdlib:
            return []
        found, areas = [], self.__dmtransfer_areas(areas)
        for area, dotsetname in itertools.product(areas, dotsetname.split('|')):
            for dotset in self.__dmdlib.group(dotsetname):
                dotset = dotset.scale(scale).asmatchcolor(color)
                res = self.__dm.FindMultiColorEx(*area, dotset.matchcolor, dotset.tmpl, similarity, 0)
                if not res: continue
                xs, ys = [[int(coor) for coor in coors.split(',')] for coors in res.split('|')]
                for i in range(min(len(xs), len(ys))):
                    x, y = xs[i], ys[i]
                    start = dotset.getrealpos((x, y))
                    end = start[0] + dotset.width, start[1] + dotset.height
                    found.append(WinTarget(*start, *end, dotsetname, similarity, self))
        return found

    def dmfindword(self, texts: str, areas: _Area|_Areas, color: str, similarity=0.9) -> WinTarget:
        while self.__dmwordlock: pass
        self.__dmwordlock = True
        found = WinTarget()
        for area in self.__dmtransfer_areas(areas):
            x, y, index = self.__dm.FindStrFast(*area, texts, color, similarity)
            if index > -1:
                found = WinTarget(x, y, x, y, texts.split('|')[index], similarity, self)
                break
        self.__dmwordlock = False
        return found

    def dmfindwords(self, texts: str, areas: _Area|_Areas, color: str, similarity=0.9) -> list[WinTarget]:
        while self.__dmwordlock: pass
        self.__dmwordlock = True
        found = []
        textList = texts.split('|')
        for area in self.__dmtransfer_areas(areas):
            res = self.__dm.FindStrFastEx(*area, texts, color, similarity)
            if not res: continue
            for info in res.split('|'):
                index, x, y = info.split(',')
                word = textList[int(index)]
                pos = int(x), int(y)
                found.append(WinTarget(*pos, *pos, word, similarity, self))
        self.__dmwordlock = False
        return found



import numpy
from typing import TypedDict, Literal

__all__ = [
    '_AppCSS',
]

_AppRunningCss = TypedDict('_AppRunningCss', {
    'cssname': Literal['normal', 'hover', 'active']
})

_AppCSS = TypedDict('_AppCSS', {
    'win-title': str,
    'win-pos': tuple[int, int] | Literal['unset'],
    'win-size': tuple[int, int] | Literal['unset'],
    'win-icon': str,
    'win-visible': bool,
    'win-opacity': float,
    'win-caption-theme': Literal['active', 'blur', 'toggle'],
    'win-ask-close': bool,
    'win-cursor': Literal['arrow', 'ibeam', 'wait', 'cross', 'uparrow', 'sizenwse', 'sizenesw', 'sizewe', 'sizens', 'sizeall', 'no', 'hand', 'appstarting', 'help'],

    'bk-color': int,

    'bk-image': numpy.ndarray | Literal['none'],
    'bk-image-pos': tuple[int, int],
    'bk-image-crop-start': tuple[int, int],
    'bk-image-crop-size': tuple[int, int] | Literal['auto'],

    'bk-text': str,
    'bk-text-color': int,
    'bk-text-align': Literal['LT', 'LM', 'LB', 'RT', 'RM', 'RB', 'MT', 'MM', 'MB'],
    'bk-text-offset': tuple[int, int],
    'bk-text-overflow': Literal['clip', 'ellipsis', 'wrap'],

    'bk-font': str,
    'bk-font-size': int,
    'bk-font-weight': Literal['normal', 'thin', 'heavy'],
    'bk-font-italic': bool,
    'bk-font-strike': bool,
    'bk-font-underline': bool,
}, total=False)



from typing import Callable
from functools import wraps
from aclib.winlib import winapi

class reactive_dict(dict):
    __old = dict
    __new = dict
    __diffkeys = str

    def __init__(self, data: dict={}, *, on_update: Callable[[__old, __new, __diffkeys], None]=None):
        super().__init__(data)
        self.__update_event = on_update

    def update(self, data: dict={}, **kwargs):
        old = self.copy()
        super().update(data, **kwargs)
        if self.__update_event:
            diffkeys = '|'.join(k for k, v in self.items() if k not in old or old[k] is not v)
            self.__update_event(old, self, diffkeys)

    def unref(self):
        self.__update_event = None

def apptask(task, task_thread=0):
    @wraps(task)
    def wrappedtask(self, *args, **kwargs):
        task_tid = task_thread or winapi.GetWindowThreadProcessId(self.handle)[0]
        if winapi.GetCurrentThreadId() == task_tid:
            return task(self, *args, **kwargs)
        return winapi.RequestMsgloopTask(task, task_tid, self, *args, **kwargs)
    return wrappedtask



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self

from aclib.winlib.winapi import *

from .basewindow import BaseWindow


__all__ = [
    'Window'
]


class Window(BaseWindow):


    @classmethod
    def fromhandle(cls, hwnd: int) -> Self:
        return cls()._sethandle_(hwnd)

    def __repr__(self):
        alias = ''
        if self.handle == GetDesktopWindow():
            alias = '.desktopwindow'
        if self.handle == GetDesktopView():
            alias = '.desktop'
        if self.handle == GetTaskbarWindow():
            alias = '.taskbar'
        return super().__repr__().replace(self.__class__.__name__, f'{self.__class__.__name__}{alias}', 1)


    @classmethod
    def findwindow(cls, title='', classname='', visible: bool=None) -> Self:
        return cls.fromhandle(FilterWindow(IterDescendantWindows(0), title, classname, visible))

    @classmethod
    def findwindows(cls, title='', classname='', visible: bool=None) -> list[Self]:
        return [cls.fromhandle(h) for h in FilterWindows(IterDescendantWindows(0), title, classname, visible)]


    @classmethod
    def desktopwindow(cls) -> Self:
        """winapi中定义的DesktopWindow"""
        return cls.fromhandle(GetDesktopWindow())

    @classmethod
    def desktop(cls) -> Self:
        """显示桌面图标的窗口"""
        return cls.fromhandle(GetDesktopView())

    @classmethod
    def taskbar(cls) -> Self:
        return cls.fromhandle(GetTaskbarWindow())


    @classmethod
    def pointwindow(cls, pos: tuple[int, int] = None) -> Self:
        return cls.fromhandle(GetPointWindow(pos or GetCursorPos()))

    @classmethod
    def foregroundwindow(cls) -> Self:
        return cls.fromhandle(GetForegroundWindow())


    def parent(self) -> Self:
        return self.fromhandle(GetParentWindow(self.handle))

    def root(self) -> Self:
        return self.fromhandle(GetRootWindow(self.handle))

    def rootowner(self) -> Self:
        return self.fromhandle(GetRootOwnerWindow(self.handle))


    def prevbrother(self) -> Self:
        return self.fromhandle(GetPrevWindow(self.handle))

    def nextbrother(self) -> Self:
        return self.fromhandle(GetNextWindow(self.handle))

    def brother(self, title='', classname='', visible: bool=None) -> Self:
        return self.fromhandle(FilterWindow(IterBrotherWindows(self.handle), title, classname, visible))

    def brothers(self, title='', classname='', visible: bool=None) -> list[Self]:
        return [self.fromhandle(h) for h in FilterWindows(IterBrotherWindows(self.handle), title, classname, visible)]


    def child(self, title='', classname='', visible: bool=None) -> Self:
        return self.fromhandle(FilterWindow(IterChildWindows(self.handle), title, classname, visible))

    def children(self, title='', classname='', visible: bool=None) -> list[Self]:
        return [self.fromhandle(h) for h in FilterWindows(IterChildWindows(self.handle), title, classname, visible)]


    def descendant(self, title='', classname='', visible: bool=None) -> Self:
        return self.fromhandle(FilterWindow(IterDescendantWindows(self.handle), title, classname, visible))

    def descendants(self, title='', classname='', visible: bool=None) -> list[Self]:
        return [self.fromhandle(h) for h in FilterWindows(IterDescendantWindows(self.handle), title, classname, visible)]



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import numpy as np
    from ._typing import _AppCSS

import ctypes
from aclib.winlib import wintype, wincon
from aclib.winlib.wincon import *
from aclib.winlib.__API__._windll import user32, gdi32


class BackgroundBuffer:

    def __init__(self):
        self._canvas = Canvas()

    def resize(self, size: tuple[int,int]):
        self._canvas.resize(size)

    def redraw(self, c: _AppCSS, refont: bool):
        self._canvas.draw_rect(c['bk-color'])
        if not isinstance(im := c['bk-image'], str):
            srcpos = c['bk-image-crop-start']
            destx, desty = c['bk-image-pos']
            destsize = c['bk-image-crop-size']
            destw, desth = im.shape[:2][::-1] if destsize == 'auto' else destsize
            imcanvas = Canvas(im.shape[:2][::-1])
            imcanvas.draw_image(im)
            imcanvas.copy_tocanvas(self._canvas, (destx, desty, destx+destw, desty+desth), srcpos)
        if refont:
            fw = getattr(wincon, f'FW_{c["bk-font-weight"].upper()}')
            self._canvas.refont(c['bk-font'], c['bk-font-size'], fw, c['bk-font-italic'], c['bk-font-underline'], c['bk-font-strike'])
        if text := c['bk-text']:
            color = c['bk-text-color']
            _dx, _dy = c['bk-text-offset']
            _overflow = c['bk-text-overflow']
            _xalign, _yalign = c['bk-text-align']
            area = (_dx, _dy, self._canvas._size[0] + _dx, self._canvas._size[1] + _dy)
            dtformat = DT_WORDBREAK if _overflow == 'wrap' else DT_SINGLELINE
            if _overflow == 'ellipsis':
                dtformat |= DT_END_ELLIPSIS
            _xalign = {'L': DT_LEFT, 'R': DT_RIGHT, 'M': DT_CENTER}[_xalign]
            _yalign = {'T': DT_TOP, 'B': DT_BOTTOM, 'M': DT_VCENTER}[_yalign]
            dtformat |= _xalign | _yalign
            self._canvas.draw_text(text, color, area, dtformat)

    def repaint(self, hwnd: int):
        ps = wintype.PAINTSTRUCT()
        hdc = user32.BeginPaint(hwnd, ps.ref)
        self._canvas.copy_todc(hdc)
        user32.EndPaint(hwnd, ps.ref)
        user32.ReleaseDC(hwnd, hdc)


class Canvas:
    def __init__(self, size = (1, 1)):
        self._size = size
        self._mem = gdi32.CreateCompatibleDC(0)
        self._bmp = self._oldbmp = 0
        self._font = self._oldfont = 0
        self.resize(size)
        gdi32.SetBkMode(self._mem, TRANSPARENT)

    def __del__(self):
        self.refont(None)
        self.resize(None)
        gdi32.DeleteDC(self._mem)

    def resize(self, size: tuple[int,int]|None):
        gdi32.SelectObject(self._mem, self._oldbmp)
        gdi32.DeleteObject(self._bmp)
        self._size = (1,1)
        if size is not None:
            self._size = size
            self._bmp = gdi32.CreateBitmap(*size, 1, 32, 0)
            self._oldbmp = gdi32.SelectObject(self._mem, self._bmp)

    def refont(self, font: str|None, size=18, weight=400, italic=False, underline=False, strikeout=False):
        gdi32.SelectObject(self._mem, self._oldfont)
        gdi32.DeleteObject(self._font)
        if font is not None:
            self._font = gdi32.CreateFontW(size, 0,0,0, weight, italic, underline, strikeout, 0,0,0,0,0, font)
            self._oldfont = gdi32.SelectObject(self._mem, self._font)

    def draw_rect(self, color: int, area: tuple[int,int,int,int] = ...):
        area  = (0, 0, *self._size) if area is ... else area
        hb = gdi32.CreateSolidBrush(color)
        user32.FillRect(self._mem, wintype.RECT(*area).ref, hb)
        gdi32.DeleteObject(hb)

    def draw_image(self, im: np.ndarray):
        bmi = wintype.BITMAPINFO()
        bmi.bmiHeader.biSize = bmi.bmiHeader.sizeof()
        bmi.bmiHeader.biWidth = im.shape[1]
        bmi.bmiHeader.biHeight = -im.shape[0]
        bmi.bmiHeader.biPlanes = 1
        bmi.bmiHeader.biBitCount = 32
        bmi.bmiHeader.biCompression = BI_RGB
        gdi32.SetDIBits(0, self._bmp, 0, im.shape[0], im.ctypes.data_as(wintype.PBYTE), bmi.ref, DIB_RGB_COLORS)

    def draw_text(self, text: str, color = 0x000000, area: tuple[int,int,int,int] = ..., format = 0):
        rect = wintype.RECT(*area if area is not ... else (0, 0, *self._size))
        gdi32.SetTextColor(self._mem, color)
        user32.DrawTextW(self._mem, text, -1, rect.ref, format)

    def copy_todc(self, tohdc: int, toarea: tuple[int,int,int,int] = ..., copystart: tuple[int,int] = (0,0)):
        area = (0, 0, *self._size) if toarea is ... else toarea
        gdi32.BitBlt(tohdc, *area[:2], *self._size, self._mem, *copystart, SRCCOPY)

    def copy_tocanvas(self, canvas: Canvas, toarea: tuple[int,int,int,int] = ..., copystart: tuple[int,int] = (0,0)):
        self.copy_todc(canvas._mem, toarea, copystart)

    def copy_towindow(self, hwnd: int, toarea: tuple[int,int,int,int] = ..., copystart: tuple[int,int] = (0,0)):
        hdc = user32.GetDC(hwnd)
        self.copy_todc(hdc, toarea, copystart)
        user32.ReleaseDC(hwnd, hdc)

    def to_bytes(self) -> bytearray:
        total_bytes = self._size[0] * self._size[1] * 4
        buffer = bytearray(total_bytes)
        byte_array = ctypes.c_ubyte * total_bytes
        gdi32.GetBitmapBits(self._bmp, total_bytes, byte_array.from_buffer(buffer))
        return buffer



from __future__ import annotations
from typing import Sequence, Self

import time
from aclib.builtins import MatTarget

from .basewindow import BaseWindow


__all__ = [
    '_Pos',
    '_Size',
    '_Area',
    '_Areas',
    'WinTarget',
    'WinTargetEx',
]


_Pos = _Size = tuple[int, int]
_Area = tuple[int, int, int, int]
_Areas = Sequence[_Area]


class WinTarget(MatTarget):

    @classmethod
    def fromtarget(cls, target: MatTarget, bindwin: BaseWindow):
        return cls(*target.border, target.name, target.similarity, bindwin)

    def __init__(
        self, left=0, top=0, right=0, bottom=0, name='', similarity=0.0,
        bindwin: BaseWindow = BaseWindow()
    ):
        super().__init__(left, top, right, bottom, name, similarity)
        self._boundwin = bindwin

    def click_if_exists(self, wait=0.0) -> Self:
        if self:
            self._boundwin.leftclick(self.center)
            time.sleep(wait)
        return self

    def toEx(self, closepos: _Pos) -> WinTargetEx:
        return WinTargetEx.fromtarget(self, self._boundwin, closepos)


class WinTargetEx(WinTarget):

    @classmethod
    def fromtarget(cls, target: MatTarget, bindwin: BaseWindow, closepos: _Pos):
        return cls(*target.border, target.name, target.similarity, bindwin, closepos)

    def __init__(
        self, left=0, top=0, right=0, bottom=0, name='', similarity=0.0,
        bindwin: BaseWindow = BaseWindow(), closepos: _Pos = (-1, -1),
    ):
        super().__init__(left, top, right, bottom, name, similarity, bindwin)
        self._closepos = closepos

    def close_if_exists(self, wait=0.0) -> Self:
        if self:
            self._boundwin.leftclick(self._closepos)
            time.sleep(wait)
        return self



from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='aclib.autowin',
    version='1.5.4',
    author='AnsChaser',
    author_email='anschaser@163.com',
    description='auto window task',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/AnsChaser/aclib.autowin',
    python_requires='>=3.6',
    install_requires=['aclib.builtins', 'aclib.winlib'],
    extras_require={
        'cv': ['aclib.cv'],
        'dm': ['aclib.dm'],
        'full': ['aclib.autowin[cv,dm]'],
        'dev': ['aclib.autowin[full]', 'aclib.pip', 'build', 'twine']
    },
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ]
)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self, Callable, Any
    from ._typing import _AppCSS, _AppRunningCss

import ctypes, inspect
from time import sleep
from typing import overload
from aclib.builtins import decorator
from aclib.winlib import winapi, wincon, wintype
from aclib.winlib.wincon import *
from aclib.winlib.__API__._windll import user32, gdi32, kernel32
from .background import BackgroundBuffer
from .pytypes import reactive_dict, apptask

from ..__API__.basewindow import BaseWindow


__all__ = [
    'AppWindow',
]

WM_APP_REDRAW = WM_APP + 0x0080


class AppWindow(BaseWindow):

    @classmethod
    def __newappwin(cls, wtype: int, wparent: Self | int, wclass: str) -> Self:
        window = cls()
        window.__create_params = (wtype, wparent, wclass)
        return window

    @decorator.instance_classmethod
    def newoverlapped(parent: Self | int = 0, wnd_class='aclib.winapp') -> Self:
        return parent.cls.__newappwin(WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX, parent.self, wnd_class)

    @decorator.instance_classmethod
    def newpopup(parent: Self | int, wnd_class='aclib.winapp') -> Self:
        return parent.cls.__newappwin(WS_POPUP|WS_MINIMIZEBOX, parent.self, wnd_class)

    @decorator.instance_classmethod
    def newchild(parent: Self | int, wnd_class='aclib.winapp') -> Self:
        return parent.cls.__newappwin(wincon.WS_CHILD, parent.self, wnd_class)

    @decorator.instance_classmethod
    def new_button(parent: Self | int) -> Self:
        ...

    @decorator.instance_classmethod
    def new_sys_static(parent: Self | int) -> Self:
        ...

    @decorator.instance_classmethod
    def new_sys_edit(parent: Self | int) -> Self:
        ...

    def __init__(self):
        super().__init__()
        self.__listener: dict[int, dict[Callable, None]] = {}
        self.__trackingmouse = False
        self.__rect_dsize = 0
        self.__sys_control = BaseWindow()
        self.__css_calculated: _AppCSS = reactive_dict(on_update=self.__on_css_using_update__)
        self.__css_normal: _AppCSS = reactive_dict(on_update=self.__on_css_base_update__)
        self.__css_hover: _AppCSS = reactive_dict(on_update=self.__on_css_base_update__)
        self.__css_active: _AppCSS = reactive_dict(on_update=self.__on_css_base_update__)
        self.__css_usename: _AppRunningCss = reactive_dict({'cssname': 'normal'}, on_update=self.__on_css_base_update__)
        self.css()

    def __repr__(self):
        wndid = self.handle or hex(id(self))
        wndsummary = self.title or self.classname if self.handle else 'UNCREATED'
        return f'<{self.__class__.__name__}-{wndid} {wndsummary}>'

    def css(self, css: _AppCSS={
        'win-title': '',
        'win-pos': 'unset',
        'win-size': 'unset',
        'win-visible': True,
        'win-opacity': 1.0,
        'win-caption-theme': 'toggle',
        'win-ask-close': False,
        'win-cursor': 'arrow',
        'win-icon': '',
        'bk-color': 0xffffff,
        'bk-image': 'none',
        'bk-image-pos': (0,0),
        'bk-image-crop-start': (0,0),
        'bk-image-crop-size': 'auto',
        'bk-text': '',
        'bk-text-color': 0x0,
        'bk-text-align': 'MM',
        'bk-text-offset': (0,0),
        'bk-text-overflow': 'clip',
        'bk-font': '微软雅黑',
        'bk-font-size': 18,
        'bk-font-weight': 'normal',
        'bk-font-italic': False,
        'bk-font-strike': False,
        'bk-font-underline': False,
    }) -> Self:
        """
        - 此方法可多次、链式地调用。不传参数，样式表将被重置为默认值；传入参数，同名属性将被更新
        - 窗口被创建后，配置 win-pos 和 win-size 只会修改子窗口的位置和大小，对于非子窗口，请使用 app.setpos 和 app.setsize
        - bk-text-overflow 设为 wrap 时，bk-text-align 指定的纵向对齐方式将无效
        - bk-color 目前仅支持十进制 BGR 颜色值，如 0xffffff
        """
        return self.__css_normal.update(css) or self

    def csshover(self, css: _AppCSS) -> Self:
        """鼠标悬停在窗口上时生效，此方法可多次、链式地调用，新的样式表将会更新原来的样式表"""
        return self.__css_hover.update(css) or self

    def cssactive(self, css: _AppCSS) -> Self:
        """鼠标在窗口上按下时生效，此方法可多次、链式地调用，新的样式表将会更新原来的样式表"""
        return self.__css_active.update(css) or self

    def cssget(self, name: str):
        return self.__css_calculated.get(name)

    def create(self) -> Self:
        assert self.handle == 0, 'AppWindow could be created only once.'
        wtype, wparent, wclass = self.__create_params
        hparent = getattr(wparent, 'handle', wparent)
        if not winapi.GetWindowClassInfo(wclass).hInstance:
            winapi.RegisterWindowClass(wclass, self.__wndproc)
        self.__checkthread()
        if wclass == 'listbox':
            wtype |= WS_VSCROLL | LBS_NOTIFY
        h = apptask(winapi.CreateWindow, self.__appthread)(
            hparent, wclass, '', 0, 0, wtype|ES_AUTOHSCROLL, wincon.WS_EX_LAYERED, id(self))
        if h and winapi.IsWindowClassInSystem(wclass):
            self._sethandle_(h)
            self._ctl_hbkclr = 0
            self._bk = BackgroundBuffer()
            self.__css_usename.update(cssname='normal')
            self.__appwindows[h] = self
            winapi.RedrawWindow(h)
        self.__checkwindow()
        return self

    def loop(self):
        while winapi.IsWindowExistent(self.handle):
            sleep(1)


    def __on_css_base_update__(self, old: _AppCSS, new: _AppCSS, _):
        if self.handle == 0:
            return
        usename = self.__css_usename['cssname']
        calculated = {} | self.__css_normal
        if usename in ['hover', 'active']:
            calculated |= self.__css_hover
        if usename == 'active':
            calculated |= self.__css_active
        self.__css_calculated.update(calculated)

    def __on_css_using_update__(self, old: _AppCSS, new: _AppCSS, diffkeys: str):
        if 'win-title' in diffkeys:
            self.settitle(new['win-title'])
            self.__sys_control.settitle(new['win-title'])
        if 'win-pos' in diffkeys or 'win-size' in diffkeys \
        and winapi.GetWindowLong(self.handle, wincon.GWL_STYLE) & wincon.WS_CHILDWINDOW:
            pos, size = new['win-pos'], new['win-size']
            pos, size = 0 if pos == 'unset' else pos, 0 if size == 'unset' else size
            self.setpossizeR(pos, size)
            self.__sys_control.setsize(size)
        if 'win-icon' in diffkeys:
            winapi.SendMessage(self.handle, wincon.WM_SETICON, 0, 0)
        if 'win-opacity' in diffkeys:
            self.settransparency(new['win-opacity'])
        if 'win-caption-theme' in diffkeys and (_wct := new['win-caption-theme']) in ('active', 'blur'):
            winapi.SendMessage(self.handle, wincon.WM_NCACTIVATE, _wct == 'active', 0)
        if 'win-cursor' in diffkeys:
            winapi.PostMessage(self.handle, wincon.WM_SETCURSOR, self.handle, winapi.MakeLong(wincon.HTCLIENT, wincon.WM_MOUSEMOVE))
        if 'bk-' in diffkeys:
            winapi.SendMessage(self.handle, WM_APP_REDRAW, 0, 'bk-font' in diffkeys)
        if 'bk-font' in diffkeys:
            winapi.PostMessage(self.handle, wincon.WM_SETFONT, self._bk._canvas._font, 1)
        if 'win-visible' in diffkeys:
            user32.ShowWindow(self.handle, new['win-visible'])
            user32.ShowWindow(self.__sys_control.handle, new['win-visible'])
        if diffkeys:
            winapi.RedrawWindow(self.handle)

    def minimize(self):
        for c in winapi.IterDescendantWindows(self.handle):
            self.__appwindows[c].__css_usename.update(cssname='normal')
        super().minimize()

    def __on_message__(self, h, m, w, l):
        if m == wincon.WM_CREATE:
            self._sethandle_(h)
            self._bk = BackgroundBuffer()
            self.__css_usename.update(cssname='normal')
        if m in [0x0133, 0x0134, 0x0135, 0x0138] and (control := self.__appwindows.get(l, None)):
            gdi32.SetTextColor(w, control.__css_calculated['bk-text-color'])
            gdi32.SetBkColor(w, control.__css_calculated['bk-color'])
            gdi32.DeleteObject(control._ctl_hbkclr)
            control._ctl_hbkclr = gdi32.CreateSolidBrush(control.__css_calculated['bk-color'])
            return control._ctl_hbkclr
        if m == wincon.WM_MOUSEHOVER:
            self.__css_usename.update(cssname='hover')
        if m == wincon.WM_MOUSELEAVE:
            self.__css_usename.update(cssname='normal')
        if m == wincon.WM_LBUTTONDOWN:
            user32.SetFocus(self.handle)
            self.capturemouse()
            self.__css_usename.update(cssname='active')
        if m == wincon.WM_LBUTTONUP:
            self.releasecapture()
            self.__css_usename.update(cssname='hover')
        if m == wincon.WM_NCCALCSIZE:
            new_rect, old_rect, old_client = ctypes.cast(l, ctypes.POINTER(wintype.NCCALCSIZE_PARAMS)).contents.rgrc
            dw, dh = old_rect.size[0] - old_client.size[0], old_rect.size[1] - old_client.size[1]
            if self.isminimized:
                self.__rect_dsize = dw, dh
            elif self.__rect_dsize:
                dw, dh = self.__rect_dsize
                self.__rect_dsize = 0
            new_clientsize = new_rect.size[0] - dw, new_rect.size[1] - dh
            self._bk.resize(new_clientsize)
            self._bk.redraw(self.__css_calculated, False)
        if m == WM_APP_REDRAW:
            self._bk.redraw(self.__css_calculated, h)
        if m == wincon.WM_PAINT and not self.classname.endswith('.wrapper'):
            self._bk.repaint(h)

        if m == wincon.WM_MOUSEMOVE and not self.__trackingmouse:
            winapi.TrackMouseEvent(h)
            self.__trackingmouse = True
        if m == wincon.WM_MOUSELEAVE:
            self.__trackingmouse = False

        if m == wincon.WM_COMMAND and (control := self.__appwindows.get(l, None)):
            for callback in control.__listener.get(-winapi.ParseLong(w)[1], []):
                callback(*(h, m, w, l)[:len(inspect.signature(callback).parameters)])
        for callback in self.__listener.get(m, []):
            callback(*(h, m, w, l)[:len(inspect.signature(callback).parameters)])

        if m == wincon.WM_SETCURSOR and w == h and winapi.ParseLong(l)[0] == wincon.HTCLIENT:
            cursor = self.__css_calculated['win-cursor']
            hcursor = cursor if isinstance(cursor, int) else getattr(wincon, f"IDC_{cursor.upper()}")
            winapi.SetCursor(hcursor)
        if m == wincon.WM_ERASEBKGND:
            return 1
        if m == wincon.WM_SETICON \
        and (hicon := user32.LoadImageW(0, self.__css_calculated['win-icon'], wincon.IMAGE_ICON, 0, 0, wincon.LR_LOADFROMFILE)):
            return user32.DestroyIcon(user32.DefWindowProcW(h, m, wincon.ICON_SMALL, hicon))
        if m == wincon.WM_NCACTIVATE \
        and winapi.ParseLong(w)[0] == {'active': 1, 'blur': 0}.get(self.__css_calculated['win-caption-theme'], -1) ^ 1:
            return 1
        if m == wincon.WM_CLOSE and self.__css_calculated['win-ask-close'] \
        and not winapi.MessageBox('确定要退出吗？', '提示', h, True, wincon.MB_YESNO) == wincon.IDYES:
            return 0

        if m == wincon.WM_NCDESTROY:
            self._bk._canvas.__del__()
            self.__listener.clear()
            self.__css_normal.unref()
            self.__css_hover.unref()
            self.__css_active.unref()
            self.__css_calculated.unref()
            self.__css_usename.unref()


    __appthread = 0
    __appwindows: dict[int, Self] = {}

    @classmethod
    def __wndproc(cls, hwnd, message, wparam, lparam):
        ret = None
        # if message == WM_PARENTNOTIFY and winapi.IsWindowClassInSystem(winapi.GetWindowClassName(lparam)):
        #     print(hwnd, message, wparam, lparam)
        #     message, hwnd, wparam, lparam = wparam, lparam, 0, 0
        if message == wincon.WM_CREATE:
            lpApp = ctypes.cast(lparam, ctypes.POINTER(wintype.CREATESTRUCTW)).contents.lpCreateParams
            cls.__appwindows[hwnd] = ctypes.cast(lpApp, ctypes.py_object).value
        if (app := cls.__appwindows.get(hwnd, None)) is not None:
            ret = app.__on_message__(hwnd, message, wparam, lparam)
        if message == wincon.WM_NCDESTROY:
            cls.__appwindows.pop(hwnd, None)
            cls.__checkwindow()
        return ret

    @classmethod
    def __checkthread(cls):
        if not cls.__appthread:
            cls.__appthread = winapi.CreateMsgloopThread()

    @classmethod
    def __checkwindow(cls):
        if not cls.__appwindows and cls.__appthread:
            cls.__appthread = winapi.DestroyMsgloopThread(cls.__appthread)


    @staticmethod
    def __parse_msg(msg: int | str):
        if isinstance(msg, str):
            _, sign, msg = msg.rpartition('-')
            msg = msg.upper() if msg.count('_') else f'WM_{msg}'.upper()
            msg = (-1) ** sign.count('-') * getattr(wincon, msg)
        return msg

    def addmsglistener(self, msg: int | str, callback: Callable[[int, int, int, int], Any]) -> Self:
        """
        - 回调函数按添加顺序执行
        - msg 为字符串时消息名称不区分大小写，WM_前缀可以省略： WM_LBUTTONUP == 0x0202 == 'lbuttonup' == 'WM_LBUTTONUP'
        - 如需处理win32预设控件的消息，整数参数请使用负数，字符串参数也请使用'-'前缀： -BM_CLICK == '-BM_CLICK'; -WM_CUT == '-cut'
        """
        msg = self.__parse_msg(msg)
        if msg not in self.__listener:
            self.__listener[msg] = {}
        self.__listener[msg][callback] = None
        return self

    @overload
    def removemsglistener(self, msg: int | str) -> Self:
        """remove all listener on msg"""
    @overload
    def removemsglistener(self, msg: int | str, callback: Callable[[int, int, int, int], Any]) -> Self:
        """remove specified listener"""
    def removemsglistener(self, msg: int | str, callback: Callable[[int, int, int, int], Any] = None) -> Self:
        msg = self.__parse_msg(msg)
        self.__listener.get(msg, {}).pop(callback, None)
        if not callback or not self.__listener.get(msg):
            self.__listener.pop(msg, None)
        return self


    @apptask
    def capturemouse(self):
        winapi.SetCapture(self.handle)

    @apptask
    def releasecapture(self):
        winapi.ReleaseCapture()
